from typing import Any
from urllib.parse import urlencode
from django.db.models.query import QuerySet
from django.http.response import HttpResponse as HttpResponse
from django.shortcuts import render, get_object_or_404, redirect
from django.urls import reverse
from django.views import generic
from django.db.models import Q

from webapp.forms import ArticleForm, SearchForm
from webapp.models import Article


class IndexRedirectView(generic.RedirectView):
    pattern_name = 'index'


class IndexView(generic.View):
    def get(self, request):
        context = {'articles': Article.objects.all()}

        return render(request, 'index.html', context)    


class IndexListView(generic.ListView):
    template_name = 'index.html'
    context_object_name = 'articles'
    model = Article
    ordering = ['-created_at']
    paginate_by = 5
    paginate_orphans = 1
    
    def get(self, request, *args, **kwargs):
        self.form = SearchForm(self.request.GET)
        self.search_value = None

        if self.form.is_valid():
            self.search_value = self.form.cleaned_data['search']

        return super().get(request, *args, **kwargs)

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=None, **kwargs)
        context['form'] = self.form
        
        if self.search_value:
            context['query_params'] = urlencode({'search': self.search_value})
        
        return context
    
    def get_queryset(self):
        qs = super().get_queryset()
        
        if self.search_value:
            query = (
                Q(title__icontains=self.search_value) 
                | Q(author__icontains=self.search_value)
            )
            qs = qs.filter(query).distinct()

        return qs


class ArticleDetailView(generic.TemplateView):
    template_name = 'articles/detail.html'

    def get_context_data(self, **kwargs):
        kwargs['article'] = get_object_or_404(Article, id=kwargs['id'])
        return super().get_context_data(**kwargs)


class ArticleCreateView(generic.FormView):
    template_name = 'articles/article.html'
    form_class = ArticleForm
    
    def form_valid(self, form):
        self.article = form.save()
        return super().form_valid(form)

    def get_success_url(self):
        return reverse('articles_detail', kwargs={'id': self.article.id})


class ArticleUpdateView(generic.FormView):
    template_name = 'articles/article.html'
    form_class = ArticleForm
    
    def dispatch(self, request, *args, **kwargs):
        self.article = self.get_object()
        return super().dispatch(request, *args, **kwargs)
    
    def get_object(self):
        return get_object_or_404(Article, id=self.kwargs.get('id'))

    def get_context_data(self, **kwargs):
        return super().get_context_data(
            **kwargs, 
            article=self.article,
            btn_text='Update',
        )
    
    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['instance'] = self.article
        return kwargs

    def form_valid(self, form):
        self.article = form.save()
        return super().form_valid(form)
    
    def get_success_url(self):
        return reverse('articles_detail', kwargs={'id': self.article.id})


def article_update_view(request, id):
    article = get_object_or_404(Article, id=id)
    
    if request.method == 'GET':
        form = ArticleForm(initial={
            'title': article.title,
            'text': article.text,
            'author': article.author,
            'tags': article.tags.all(),
        })

        return render(
            request, 'articles/article.html', 
            context={'article': article, 'form': form},
        )
    
    elif request.method == 'POST':
        form = ArticleForm(data=request.POST)
        
        if form.is_valid():
            tags = form.cleaned_data.pop('tags')

            article.title = form.cleaned_data.get('title')
            article.author = form.cleaned_data.get('author')
            article.text = form.cleaned_data.get('text')

            article.save()
            
            article.tags.set(tags)

            return redirect('index')
        else:
            return render(request, 'articles/article.html', context={'form': form, 'article': article})


def article_delete_view(request, id):
    get_object_or_404(Article, id=id).delete()
    
    return redirect('index')
